/*
**> Call/Put on rainbow
** 
** Let S1(t),...,Sn(t) be the vector of n assets
** We sort the vector in ascending order and obtain the X1(t),...Xn(t) with:
** 1. X1(t) = min(S1(t)/S1(0),...,Sn(t)/Sn(0))
** 2. X2(t) the second worst performed 
** etc...
** n. Xn(t) = max(S1(t)/S1(0),...,Sn(t)/Sn(0))
**
** Let Y(t) = w1*X1(t) + ... + wn*Xn(t)
** The payoff of the Call option is: max(Y(T) - K)
*/

new;
library option,pgraph;

cls;

let S0 = 100 100 100 100;            
let mu = 0.05 0.06 0.07 0.08;
let sigma = 0.15 0.25 0.20 0.30;                  // Implied Volatility

// Correlation Matrix between Asset returns

let rho = {1,
           0.5, 1,
           0.25, 0.15, 1,
           0.5,  0.5, 0.00, 1.00};
rho = xpnd(rho); 

// First, we have to correct the drift
r = 0.05;                                    // risk-free interest rate
mu = ones(rows(mu),1) * r;

// Caracteristics of the option
t = 1;                                       // 1 Fixing Date: 1 Year
tau = t[1];                                  // Maturity of the option = 1 Year
Notional = 1000;                             // Notional of the option
K = 0.03;                                    // Strike of the option = 3%
let w = 0.00 0.10 0.40 0.50;                 // Weights of the rainbow structure
fn call_payoff(S,K) = (S - K) .* (S .>= K);  // call option
fn put_payoff(S,K) = (K - S) .* (S .<= K);   // put option

Ns = 100000;  // Number of simulations
S = array_simulate_mGBM(S0,mu,sigma,rho,t,Ns);
S = mGBM_getDate(t,S,1)';
Stilde = S ./ S0 - 1;                        // We compute returns

Nx = rows(S0);
X = zeros(Nx,Ns);

i = 1;
do until i > Ns;                               // for each simulation
  X[.,i] = sortc(Stilde[.,i],1);               // we sort the assets by their performance
  i = i + 1;
endo;

Rainbow = sumc(w .* X);

C = Notional * exp(-r*tau) * meanc( call_payoff(Rainbow,K) );
P = Notional * exp(-r*tau) * meanc( put_payoff(Rainbow,K) );

print ftos(C,"Call Rainbow Option Price = %lf",3,2);
print ftos(P,"Put Rainbow Option Price = %lf",3,2);