/*
**> Griddy Gibbs Sampling
** 
*/

new;
library tsm,qam,cml,optmum,pgraph;
cls;

#include stochastic_volatility.src

datPathName = "..\\data\\";
datFileName = datPathName $+ "qam_sp500_monthly";

data = loadd(datFileName);
vnames = "" $+ getname(datFileName);
Dates = data[.,1]/1e6;

let string selected_Names = "AdjClose";
indx_Selected_Names = indsav(selected_Names,vnames);
data = data[.,indx_Selected_Names];

BeginDate = 19500101;
EndDate = 20091231; // EndDate = 20100201;

data = selif(Dates~data, (Dates .>= BeginDate) .and (Dates .<= EndDate));
Dates = data[.,1]; data = data[.,2];
nDates = rows(Data);

Rdt = Data./lag1(Data)-1;
Rdt = packr(Rdt);
T = rows(Rdt);

p = 1; q = 1;
{theta_Garch,cl_Garch,cov_Garch,logl_Garch} = regGarch(Rdt,1,p,q,0);
h_Garch = error(0) | _garch_h;
res_Garch = error(0) | Rdt - theta_Garch[1];
res_Garch = res_Garch ./ h_Garch;

Nr = 1000;       // Burn-in sample
Ns = 10000;      // Number of simulations

// hyperparameters

// beta
beta0 = 0;
A0 = 9;
invA0 = invpd(A0);

// omega
omega0 = 0|0;
G0 = diagrv(eye(2),100|100);
invG0 = invpd(G0);
nu0 = 10; 
sigma2v = 0.15^2; 
s20 = (nu0-2.0)/nu0 * sigma2v;

// h0
m0 = 0; 
C0 = 10;
invC0 = invpd(C0);

// Storage
nobs = rows(Rdt);
N = Nr + Ns;
h_stored = miss(zeros(nobs,N),0);
beta_stored = miss(zeros(N,1),0);
gamma_stored =  miss(zeros(N,1),0);
phi_stored =  miss(zeros(N,1),0);
sigma2v_stored = miss(zeros(N,1),0);

// initialization

rndseed 12345;
h =  log(packr(h_Garch));
h0 = 0; 
beta = 0;
omega = -0.5|0.8; 

// sampler parameters

nGrid = 500;
gridStep = 0.001;

t_start = hsec;

i = 1;
do until i > N;

  gamma_ = omega[1];
  phi = omega[2];

  // Sample from h --> Griddy Gibbs sampler
  theta = omega | sigma2v;
  h = SV_GriddyGibbs(h,h0,&condProb,Rdt,beta,theta,gridStep,nGrid);
  h1 = h[1];

  // Sample h0
  invC1 = invC0 + phi^2/sigma2v;
  C1 = invpd(invC1);
  m1 = C1 * (invC0 * m0 + phi*(h1-gamma_)/sigma2v);
  h0 = m1 + sqrt(C1) * rndn(1,1);

  // Sample omega and sigma_v^2
  {omega,sigma2v} = SV_Update_Parameters(h,h0,omega0,invG0,nu0,s20);

  gamma_ = omega[1];
  phi = omega[2];

  // sample from beta distribution
  X = exp(-0.5 .* h);
  Rdt_star = Rdt .* X;
  invA1 = invA0 + X'X;
  A1 = invpd(invA1);
  beta_hat = invpd(X'X)*X'Rdt_star;
  beta1 = A1 * ( (X'X) * beta_hat + invA0 * beta0); // beta1 = A1 * ( sumc(exp(-h) .* Rdt) + invA0 * beta0);
  beta1 = A1 * ( sumc(exp(-0.5 * h) .* Rdt) + invA0 * beta0); // Pourquoi exp(-0.5*h) ?

  beta = beta1 + sqrt(A1) * rndn(1,1);

  // storage
  h_stored[.,i] = h;
  beta_stored[i] = beta;
  gamma_stored[i] = gamma_;
  phi_stored[i] = phi;
  sigma2v_stored[i] = sigma2v;

  i = i + 1;
endo;

t_end = hsec;
elapsed_time = (t_end - t_start) / 100;

ct = ftos(elapsed_time,"Computational Time = %lf ",3,2) $+ "seconds";
print ct;

all_results = vput(0,h_stored,"h");
all_results = vput(all_results,beta_stored,"beta");
all_results = vput(all_results,gamma_stored,"gamma");
all_results = vput(all_results,phi_stored,"phi");
all_results = vput(all_results,sigma2v_stored,"sigma2v");
save sv2 = all_results;

proc (1) = condProb(t,h,h0,r,beta,theta,h_grid);
  local gamma_,phi,sigma2v,nobs,n,e,ht,rt,ht_plus,ht_minus,mu_t,sigma_t,f2,f1;

  gamma_ = theta[1];
  phi = theta[2];
  sigma2v = theta[3];

  nobs = rows(h);
  n = rows(h_grid);
  e = ones(n,1);

  ht = h_grid;
  rt = r[t];

  if (t /= nobs);
    ht_plus = h[t+1] * e;
    if (t /= 1);
      ht_minus = h[t-1] * e;
    else;
      ht_minus = h0 * e;
    endif;
    mu_t = (1-phi)/(1+phi^2)*gamma_ + phi/(1+phi^2) * (ht_plus + ht_minus);
    sigma_t = sqrt(sigma2v / (1+phi^2));
  else; 
    ht_minus = h[t-1] * e;
    mu_t = gamma_ + phi .* ht_minus;
    sigma_t = sqrt(sigma2v);    
  endif;

  f1 = pdfn((rt-beta)./exp(ht/2))./exp(ht/2);
  f2 = pdfn((ht-mu_t )./sigma_t)./sigma_t ;

  retp(f1.*f2);
endp;
