/*
**> Particle Filter 
** 
*/

new;
library tsm,qam,cml,optmum,pgraph,pf;
cls;

#include stochastic_volatility.src

datPathName = "..\\data\\";
datFileName = datPathName $+ "qam_sp500_monthly";

data = loadd(datFileName);
vnames = "" $+ getname(datFileName);
Dates = data[.,1]/1e6;

let string selected_Names = "AdjClose";
indx_Selected_Names = indsav(selected_Names,vnames);
data = data[.,indx_Selected_Names];

BeginDate = 19500101;
EndDate = 20091231; // EndDate = 20100201;

data = selif(Dates~data, (Dates .>= BeginDate) .and (Dates .<= EndDate));
Dates = data[.,1]; data = data[.,2];
nDates = rows(Data);

Rdt = Data./lag1(Data)-1;
Rdt = packr(Rdt);
T = rows(Rdt);

p = 1; q = 1;
{theta_Garch,cl_Garch,cov_Garch,logl_Garch} = regGarch(Rdt,1,p,q,0);
h_Garch = error(0) | _garch_h;
res_Garch = error(0) | Rdt - theta_Garch[1];
res_Garch = res_Garch ./ h_Garch;

// hyperparameters

// beta
A0 = 0.1;

// omega
omega0 = -0.7|0.9;
G0 = diagrv(eye(2),0.025|0.25);
nu0 = 3; 
sigma2v = 0.09^2; 
s20 = (nu0-2.0)/nu0 * sigma2v;

// h0
m0 = 0; 
C0 = 0.1;

// initialization
rndseed 98244; 

/*================================================================================*/
//      Definition of the LW Particle filter 

proc prior_pdf(t,nx,x);
  local theta,y;
  theta = _pf_theta;
  theta[4] = exp(theta[4]); 
  y = (nx - theta[2] - theta[3]*x) ./ theta[4];
  retp(pdfn(y) ./ theta[4]);
endp;

proc importance_pdf(t,nx,x,z);
  retp(prior_pdf(t,nx,x));
endp;

proc likelihood_pdf(t,x,z);
  local vol,y,theta;
  theta = _pf_theta;
  vol = exp(x/2);
  y = (z - theta[1]) ./ vol;
  retp(pdfn(y) ./ vol);
endp;

proc prior_sampling(t,x);
  local theta;
  theta = _pf_theta;
  theta[4] = exp(theta[4]); 
  retp(rndmvn(theta[2] + theta[3]*x, theta[4], rows(x)));
endp;

proc importance_sampling(t,x,z);
  retp(prior_sampling(t,x));
endp;

proc initial_sampling(nS);
  retp(sqrt(C0)*rndn(Ns,1) + m0);
endp;

proc Ft(t,x);
  local theta;
  theta = _pf_theta;
  retp(theta[2] + theta[3] * x);
endp;

fn Qt(t,x) = exp(_pf_theta[4]);
fn Ht(t,x) = _pf_theta[1] + exp(x/2)*rndn(1,1);
fn Rt(t,x) = 0;

// Liu-West Particle Filter

call Particle_Filter_Set(&Ft,&Qt,&Ht,&Rt,&importance_pdf,&prior_pdf,&likelihood_pdf,0,
                                &importance_sampling,&prior_sampling,&initial_sampling);

omega = rndmvn(omega0,G0,1)';

Rdt_mean = meanc(Rdt);
theta_sv = Rdt_mean|omega|nu0*s20/(2*rndgam(1,1,nu0/2)); 
_pf_theta = theta_sv;
_pf_LW_initial_Vartheta = sqrt(A0)|sqrt(diag(G0))|1;

discount_factor = .99;
Ns = 15000;
_pf_Resampling_Threshold = round(2/3*Ns);

/*================================================================================*/

t_start = date;

{h_stored,w_stored,m,cov,theta_stored,m_theta} = LW_Particle_Filter(Rdt,Ns,theta_sv,discount_factor);

t_end = date;
et = etstr(ethsec(t_start,t_end));

print ("Computational Time = " $+  et);

nobs = rows(Rdt);

let alpha = 0.05 0.95;
nAlpha = rows(alpha);
sVol_q = zeros(nobs,nAlpha);
sVol_m = exp(0.5 * m);

t = 1;
do until t > nobs;
  w_t = w_stored[t,.]';
  h_t = array3_getmatrix(h_stored,1,t);
  sVol_q[t,.] = quantile_w(h_t,w_t,alpha)';
  t = t + 1;
endo;

hVol = _Historical_Volatility(Rdt,12);
sVol_q = exp(0.5 * sVol_q);

data = hVol~sVol_m~sVol_q;
data[.,1] = lagn(data[.,1],6);
data = 100*sqrt(12) * data;

graphset;
  window3u;

  _pdate = ""; _pframe = 0; _pnum = 2;
  _plwidth = 8; _pltype = 6; _pcolor = 9|12;
  _ptitlht = 0.25; _pnumht = 0.20; _paxht = 0.25;

  xtics(1,nDates,12*10,1);
  let xlab = "50" "60" "70" "80" "90" "2000" "10";
  asclabel(xlab,0);

  ytics(5,30,5,2);

  setwind(1);
    _pcolor = 4;
    title("\216Volatilite historique");
    _pmsgstr = "-";
    _pmsgctl = 4.425~6.45~0.07~45~2~0~5 ;
    xy(0,data[.,1]);

  setwind(2);
    _pcolor = 1;
    title("\216Volatilite stochastique\L\216Moyenne");
    _pmsgctl = 4.15~6.42~0.08~45~2~0~5 ;
    xy(0,data[.,2]);

  setwind(3);
    _pcolor = 2; _plwidth = 5; _pltype = 6;
    ytics(0,50,10,2);
    title("\216Volatilite stochastique\L\216Intervalle de confiance 90%");
    e = ones(rows(data),1);
    x = seqa(1,1,rows(data));
    _pline = (1~6).*e~x~data[.,3]~x~data[.,4]~e.*(1~2~2);
    xy(0,data[.,3:4]);

    graphprt("-c=1 -cf=sv9-1.eps");

  endwind;

t = nobs;
theta_t = array3_getmatrix(theta_stored,1,t);
w_t = w_stored[t,.]';
{theta,w_t,parents}= pf_resample(theta_t,w_t,nS);
theta[.,1] = theta[.,1];
theta[.,4] = sqrt(exp(theta[.,4]));

beta = theta[.,1];
gamma_ = theta[.,2];
phi = theta[.,3];
sigma_v = theta[.,4];

x_beta = seqa(0.0,0.02/100,101);
x_gamma = seqa(-1.4,1.20/100,101);
x_phi = seqa(0.8,0.2/100,101);
x_sigma = seqa(0.20,0.30/100,101);

{d_beta,F_beta} = regKernelDensity(beta,x_beta);
{d_gamma,F_gamma} = regKernelDensity(gamma_,x_gamma);
{d_phi,F_phi} = regKernelDensity(phi,x_phi);
{d_sigma,F_sigma} = regKernelDensity(sigma_v,x_sigma);

graphset;
  begwind;
  window(2,2,0);

  fonts("simplex simgrma");
  _pdate = ""; _pframe = 0; _pnum = 2;
  _plwidth = 10; _pltype = 6; 
  _ptitlht = 0.30; _pnumht = 0.20; _paxht = 0.25;

  setwind(1);
    _pcolor = 1;
    title("\216\202b\201 (x 10[3])");
    xtics(2,14,2,1);
    ytics(0,300,100,2);
    xy(1e3*x_beta,d_beta);

  setwind(2);
    _pcolor = 2;
    title("\216\202g\201");
    xtics(-1.2,-0.4,0.2,1);
    ytics(0,4,1,2);
    xy(x_gamma,d_gamma);

  setwind(3);
    _pcolor = 3;
    title("\216\202F\201");
    xtics(0.80,0.95,0.05,1);
    ytics(0,25,5,2);
    xy(x_phi,d_phi);

  setwind(4);
    _pcolor = 4;
    title("\216\202s\201]v[ (x 10[2])");
    xtics(20,50,5,1);
    ytics(0,10,2,2);
    xy(1e2*x_sigma,d_sigma);

    graphprt("-c=1 -cf=sv9-2.eps");

  endwind;

let alpha = 0.025 0.50 0.975;
theta = theta .* (1e3~1~1~100);
q = quantile(theta,alpha);
q = reshape(q',1,12);
print ftosa(q,"%lf",4,2);

proc quantile_w(x,w,e);
  local nx,nw,parents;
  {nx,nw,parents}= pf_resample(x,w,rows(x));
  retp(quantile(nx,e));
endp;

proc _Historical_Volatility(x,Period);
  local n,vol,i;

  n = rows(x);
  vol = miss(zeros(n,1),0);
  i = 1;
  do until i > n-Period;
    vol[i] = stdc(x[i:i+Period]);
    i = i + 1;
  endo;

  retp(vol);
endp;

/*
**

            5.56             8.24            11.03            -1.01            -0.79            -0.57             0.84             0.88             0.91            26.08            32.84            41.82 
**
*/

